import os

QHUB_HOSTNAME = os.environ.get("QHUB_HOSTNAME", "github-actions.qhub.dev")
GATEWAY_ENDPOINT = "gateway"

KEYCLOAK_USERNAME = os.environ["KEYCLOAK_USERNAME"]
KEYCLOAK_PASSWORD = os.environ["KEYCLOAK_PASSWORD"]
