TERRAFORM_VERSION = "1.0.5"

HIGHEST_SUPPORTED_K8S_VERSION = "1.23.5"

DEFAULT_QHUB_DASK_VERSION = "0.4.3"
DEFAULT_QHUB_IMAGE_TAG = "main"

DEFAULT_CONDA_STORE_IMAGE_TAG = "v0.4.11"

LATEST_SUPPORTED_PYTHON_VERSION = "3.10"
